"""Benchmark cdist and pdist functions."""

from __future__ import annotations

import typing
from functools import partial

import scipy.spatial.distance as scipy_distance
import utils  # type: ignore[import]
from abd_distances import simd

if typing.TYPE_CHECKING:
    import numpy

CAR = 200
DIM = 500
NUM_RUNS = 100

METRICS = ["euclidean", "sqeuclidean", "cosine"]
GEN_DATA = {
    "f32": utils.data_f32,
    "f64": utils.data_f64,
}


def bench_func(
    func: typing.Callable[[numpy.ndarray, numpy.ndarray], numpy.ndarray],
    gen_data_x: typing.Callable[[int, int], numpy.ndarray],
    gen_data_y: typing.Callable[[int, int], numpy.ndarray] | None,
) -> None:
    """Benchmark a distance function."""
    data_x = gen_data_x(CAR, DIM)
    if gen_data_y is None:
        for _ in range(NUM_RUNS):
            func(data_x)  # type: ignore[call-arg]
    else:
        data_y = gen_data_y(CAR, DIM)
        for _ in range(NUM_RUNS):
            func(data_x, data_y)


__benchmarks__ = [
    (
        partial(bench_func, partial(scipy_distance.cdist, metric=name), gen_data, gen_data),
        partial(bench_func, partial(simd.cdist, metric=name), gen_data, gen_data),
        f"SIMD, cdist, {name}, {dtype}",
    )
    for dtype, gen_data in GEN_DATA.items()
    for name in METRICS
] + [
    (
        partial(bench_func, partial(scipy_distance.pdist, metric=name), gen_data, None),
        partial(bench_func, partial(simd.pdist, metric=name), gen_data, None),
        f"SIMD, pdist, {name}, {dtype}",
    )
    for dtype, gen_data in GEN_DATA.items()
    for name in METRICS
]



"""SIMD accelerated distance functions."""

from .abd_distances import simd

euclidean = simd.euclidean
sqeuclidean = simd.sqeuclidean
cosine = simd.cosine
cdist = simd.cdist
pdist = simd.pdist


__all__ = [
    "euclidean",
    "sqeuclidean",
    "cosine",
    "cdist",
    "pdist",
]



"""Tests that the version is correct and that the module is importable."""

import inspect

import symagen


def test_version() -> None:
    assert symagen.__version__ == "0.3.0"


def test_module() -> None:
    modules = [
        name for name, _ in inspect.getmembers(symagen) if not name.startswith("_")
    ]
    assert "guess_the_number" in modules



"""SyMaGen: Synthetic Manifold Generation."""

from .symagen import *  # noqa: F403

__doc__ = symagen.__doc__  # type: ignore[name-defined]  # noqa: F405, A001
if hasattr(symagen, "__all__"):  # type: ignore[name-defined]  # noqa: F405
    __all__ = symagen.__all__  # type: ignore[name-defined]   # noqa: F405

__version__ = "0.3.0"



"""Tests for the SIMD-accelerated distance functions."""

import math
import typing
from functools import partial

import numpy
import scipy.spatial.distance as scipy_distance
from abd_distances import vectors as abd_distances

Functions = tuple[
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
]


FUNCTIONS: dict[str, Functions] = {
    "BrayCurtis": (abd_distances.braycurtis, scipy_distance.braycurtis),
    "Canberra": (abd_distances.canberra, scipy_distance.canberra),
    "Chebyshev": (abd_distances.chebyshev, scipy_distance.chebyshev),
    "Euclidean": (abd_distances.euclidean, scipy_distance.euclidean),
    "SqEuclidean": (abd_distances.sqeuclidean, scipy_distance.sqeuclidean),
    "Manhattan": (abd_distances.manhattan, scipy_distance.cityblock),
    "Cosine": (abd_distances.cosine, scipy_distance.cosine),
}


METRICS = [
    "braycurtis",
    "canberra",
    "chebyshev",
    "euclidean",
    "sqeuclidean",
    "cityblock",
    "cosine",
]


def test_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for a in data_f32:
        for b in data_f32:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                for name, (simd_func, scipy_func) in FUNCTIONS.items():
                    _check_distances(a_, b_, name, simd_func, scipy_func, 1e-5)


def test_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for a in data_f64:
        for b in data_f64:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                for name, (simd_func, scipy_func) in FUNCTIONS.items():
                    _check_distances(a_, b_, name, simd_func, scipy_func, 1e-10)


def test_minkowski_f32(data_f32: numpy.ndarray):
    """Test the Minkowski distance function."""
    for p in range(3, 7):
        abd_mink = partial(abd_distances.minkowski, p=p)
        scipy_mink = partial(scipy_distance.minkowski, p=p)
        for a in data_f32:
            for b in data_f32:
                for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                    _check_distances(a_, b_, f"Minkowski, p={p}", abd_mink, scipy_mink, 1e-5)


def test_minkowski_f64(data_f64: numpy.ndarray):
    """Test the Minkowski distance function."""
    for p in range(3, 7):
        abd_mink = partial(abd_distances.minkowski, p=p)
        scipy_mink = partial(scipy_distance.minkowski, p=p)
        for a in data_f64:
            for b in data_f64:
                for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                    _check_distances(a_, b_, f"Minkowski, p={p}", abd_mink, scipy_mink, 1e-10)


def test_cdist_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _cdist_helper(
            data_f32,
            data_f32,
            metric,
            1e-5,
        )


def test_cdist_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _cdist_helper(
            data_f64,
            data_f64,
            metric,
            1e-10,
        )


def test_pdist_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _pdist_helper(
            data_f32,
            metric,
            1e-5,
        )


def test_pdist_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _pdist_helper(
            data_f64,
            metric,
            1e-10,
        )


def _check_distances(  # noqa: PLR0913
    a: numpy.ndarray,
    b: numpy.ndarray,
    name: str,
    abd_func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    scipy_func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    rel_tol: float,
) -> None:
    """Helper function for running tests."""
    dist = abd_func(a, b)
    expected = scipy_func(a, b)
    if "cosine" in name.lower():
        # The `scipy` implementation of the cosine distance has a lot of numerical
        # instability. So we use a small absolute tolerance instead of a relative one.
        assert math.isclose(
            dist,
            expected,
            abs_tol=1e-4,
        ), f"{name}: Expected: {expected:.8e}, got: {dist:.8e}, abs_tol: {1e-4:.8e}"
    else:
        assert math.isclose(
            dist,
            expected,
            rel_tol=rel_tol,
        ), f"{name}: Expected: {expected:.8e}, got: {dist:.8e}, rel_tol: {rel_tol:.8e}"


def _cdist_helper(
    a: numpy.ndarray,
    b: numpy.ndarray,
    metric: str,
    rel_tol: float,
) -> None:
    """Helper function for the SIMD-accelerated distance functions."""
    distances = abd_distances.cdist(a, b, metric)
    assert distances.shape == (a.shape[0], b.shape[0])
    expected = scipy_distance.cdist(a, b, metric)
    if "cosine" in metric.lower():
        # The `scipy` implementation of the cosine distance has a lot of numerical
        # instability. So we use a small absolute tolerance instead of a relative one.
        assert numpy.allclose(
            distances,
            expected,
            atol=1e-4,
        ), f"{metric}: Expected: {expected}, got: {distances}, abs_tol: {1e-4:.8e}"
    else:
        assert numpy.allclose(
            distances,
            expected,
            rtol=rel_tol,
        ), f"{metric}: Expected: {expected}, got: {distances}, rel_tol: {rel_tol:.8e}"


def _pdist_helper(
    a: numpy.ndarray,
    metric: str,
    rel_tol: float,
) -> None:
    """Helper function for the SIMD-accelerated distance functions."""
    distances = abd_distances.pdist(a, metric)
    num_distances = (a.shape[0] * (a.shape[0] - 1)) // 2
    assert distances.shape == (num_distances,)
    expected = scipy_distance.pdist(a, metric)
    if "cosine" in metric.lower():
        # The `scipy` implementation of the cosine distance has a lot of numerical
        # instability. So we use a small absolute tolerance instead of a relative one.
        assert numpy.allclose(
            distances,
            expected,
            atol=1e-4,
        ), f"{metric}: Expected: {expected}, got: {distances}, abs_tol: {1e-4:.8e}"
    else:
        assert numpy.allclose(
            distances,
            expected,
            rtol=rel_tol,
        ), f"{metric}: Expected: {expected}, got: {distances}, rel_tol: {rel_tol:.8e}"



import numpy

def euclidean(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Compute the Euclidean distance between two vectors."""

def sqeuclidean(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Compute the squared Euclidean distance between two vectors."""

def cosine(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Compute the cosine similarity between two vectors."""

def cdist(a: numpy.ndarray, b: numpy.ndarray, metric: str) -> numpy.ndarray:
    """Compute the distance between each pair of the two collections of vectors."""

def pdist(a: numpy.ndarray, metric: str) -> numpy.ndarray:
    """Compute the distance between each pair of the collection of vectors."""



"""Pytest configuration file."""

import numpy
import pytest

CAR = 20
MAX_DIM = 4
PARAMS = [10**i for i in range(1, MAX_DIM + 1)]
IDS = [f"10^{i}" for i in range(1, MAX_DIM + 1)]


def gen_data(car: int, dim: int) -> numpy.ndarray:
    """Generate random data."""
    rng = numpy.random.default_rng()
    return rng.random((car, dim))


@pytest.fixture(params=PARAMS, ids=IDS)
def data_f32(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    return gen_data(CAR, dim).astype(numpy.float32)


@pytest.fixture(params=PARAMS, ids=IDS)
def data_f64(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    return gen_data(CAR, dim)


@pytest.fixture(params=PARAMS, ids=IDS)
def data_i32(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    data = (gen_data(CAR, dim) > 0.5).astype(numpy.int32)
    data[data == 0] = -1
    return data


@pytest.fixture(params=PARAMS, ids=IDS)
def data_i64(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    data = (gen_data(CAR, dim) > 0.5).astype(numpy.int64)
    data[data == 0] = -1
    return data


@pytest.fixture(params=PARAMS, ids=IDS)
def data_u32(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    return (gen_data(CAR, dim) > 0.5).astype(numpy.uint32)


@pytest.fixture(params=PARAMS, ids=IDS)
def data_u64(request: pytest.FixtureRequest) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    dim: int = request.param
    return (gen_data(CAR, dim) > 0.5).astype(numpy.uint64)


@pytest.fixture(params=PARAMS[:-2], ids=IDS[:-2])
def strings(request: pytest.FixtureRequest) -> list[str]:
    """Return a list of strings."""
    alphabet = "ACTGN"
    length: int = request.param
    rng = numpy.random.default_rng()
    return ["".join(rng.choice(list(alphabet), length)) for _ in range(CAR)]



"""Type stubs for vector distances."""

import numpy

def braycurtis(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Bray-Curtis distance between two vectors."""

def canberra(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Canberra distance between two vectors."""

def chebyshev(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Chebyshev distance between two vectors."""

def euclidean(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Euclidean distance between two vectors."""

def sqeuclidean(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Squared Euclidean distance between two vectors."""

def manhattan(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Manhattan distance between two vectors."""

def minkowski(a: numpy.ndarray, b: numpy.ndarray, p: float) -> float:
    """Minkowski distance between two vectors."""

def cosine(a: numpy.ndarray, b: numpy.ndarray) -> float:
    """Cosine similarity between two vectors."""

def cdist(
    a: numpy.ndarray,
    b: numpy.ndarray,
    metric: str,
    p: int | None = None,
) -> numpy.ndarray:
    """Compute distance between each pair of the two collections of inputs.

    `p` is only used for Minkowski distance.
    """

def pdist(
    a: numpy.ndarray,
    metric: str,
    p: int | None = None,
) -> numpy.ndarray:
    """Pairwise distances between observations in n-dimensional space.

    `p` is only used for Minkowski distance.
    """



"""Tests for the string distance functions."""

import editdistance
from abd_distances import strings as abd_strings


def test_hamming(strings: list[str]):
    """Test the Hamming distance function."""
    assert abd_strings.hamming("NAJIBEATSPEPPERS", "NAJIBPEPPERSEATS") == 10
    assert abd_strings.hamming("TOMEATSWHATFOODEATS", "FOODEATSWHATTOMEATS") == 13

    for a in strings:
        for b in strings:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                dist = abd_strings.hamming(a_, b_)
                expected = sum(1 for i in range(len(a_)) if a_[i] != b_[i])
                assert dist == expected, f"Expected: {expected}, got: {dist}"


def test_levenshtein(strings: list[str]):
    """Test the Levenshtein distance function."""
    assert abd_strings.levenshtein("NAJIBEATSPEPPERS", "NAJIBPEPPERSEATS") == 8
    assert abd_strings.levenshtein("TOMEATSWHATFOODEATS", "FOODEATSWHATTOMEATS") == 6

    for a in strings:
        for b in strings:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                dist = abd_strings.levenshtein(a_, b_)
                expected = editdistance.eval(a_, b_)
                assert dist == expected, f"Expected: {expected}, got: {dist}"


def test_needleman_wunsch(strings: list[str]):
    """Test the Needleman-Wunsch distance function."""
    assert abd_strings.needleman_wunsch("NAJIBEATSPEPPERS", "NAJIBPEPPERSEATS") == 8
    assert abd_strings.needleman_wunsch("TOMEATSWHATFOODEATS", "FOODEATSWHATTOMEATS") == 6

    for a in strings:
        for b in strings:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                dist = abd_strings.needleman_wunsch(a_, b_)
                expected = editdistance.eval(a_, b_)
                assert dist == expected, f"Expected: {expected}, got: {dist}"



"""Benchmark vector distances."""

from __future__ import annotations

import typing
from functools import partial

import scipy.spatial.distance as scipy_distance
import utils  # type: ignore[import]
from abd_distances import vectors

if typing.TYPE_CHECKING:
    import numpy


CAR = 20
DIM = 500
NUM_RUNS = 100

GEN_DATA = {
    "f32": utils.data_f32,
    "f64": utils.data_f64,
}

METRICS = [
    "braycurtis",
    "canberra",
    "chebyshev",
    "euclidean",
    "sqeuclidean",
    "cityblock",
    "cosine",
]

FUNCTIONS = {
    f"{name}, {dtype}": (
        getattr(scipy_distance, name),
        getattr(vectors, name),
        gen_data,
    )
    for name in METRICS
    for dtype, gen_data in GEN_DATA.items()
}


def bench_func(
    func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    gen_data: typing.Callable[[int, int], numpy.ndarray],
) -> None:
    """Benchmark a distance function."""
    data_x = gen_data(CAR, DIM)
    data_y = gen_data(CAR, DIM)
    for _ in range(NUM_RUNS):
        for a in data_x:
            for b in data_y:
                _ = func(a, b)


PAIR_METRICS = [
    "cdist",
    "pdist",
]

PAIR_FUNCTIONS = {
    f"{name}, {metric}, {dtype}": (
        getattr(scipy_distance, name),
        getattr(vectors, name),
        metric,
        gen_data_x,
        gen_data_x if name == "cdist" else None,
    )
    for name in PAIR_METRICS
    for dtype, gen_data_x in GEN_DATA.items()
    for metric in METRICS
    if "minkowski" not in metric
}


def bench_pair_func(
    func: typing.Callable[[numpy.ndarray, numpy.ndarray], numpy.ndarray],
    metric: str,
    gen_data_x: typing.Callable[[int, int], numpy.ndarray],
    gen_data_y: typing.Callable[[int, int], numpy.ndarray] | None,
) -> None:
    """Benchmark a pair distance function."""
    data_x = gen_data_x(CAR * 10, DIM)

    if gen_data_y is None:
        for _ in range(NUM_RUNS):
            _ = func(data_x, metric=metric)  # type: ignore[call-arg]
        return

    data_y = gen_data_y(CAR * 10, DIM)
    for _ in range(NUM_RUNS):
        _ = func(data_x, data_y, metric=metric)  # type: ignore[call-arg]


__benchmarks__ = [
    (
        partial(bench_func, scipy_func, gen_data),
        partial(bench_func, abd_func, gen_data),
        name,
    )
    for name, (scipy_func, abd_func, gen_data) in FUNCTIONS.items()
] + [
    (
        partial(bench_pair_func, scipy_func, metric, gen_data_x, gen_data_y),
        partial(bench_pair_func, abd_func, metric, gen_data_x, gen_data_y),
        name,
    )
    for name, (
        scipy_func,
        abd_func,
        metric,
        gen_data_x,
        gen_data_y,
    ) in PAIR_FUNCTIONS.items()
]



"""Distances and similarity measures for vectors."""

from .abd_distances import vectors

braycurtis = vectors.braycurtis
canberra = vectors.canberra
chebyshev = vectors.chebyshev
euclidean = vectors.euclidean
sqeuclidean = vectors.sqeuclidean
manhattan = vectors.manhattan
minkowski = vectors.minkowski
cosine = vectors.cosine
cdist = vectors.cdist
pdist = vectors.pdist


__all__ = [
    "braycurtis",
    "canberra",
    "chebyshev",
    "euclidean",
    "sqeuclidean",
    "manhattan",
    "minkowski",
    "cosine",
    "cdist",
    "pdist",
]



"""Plotting utilities for the benchmarks."""

from __future__ import annotations

import pathlib

import numpy
import seaborn

IMAGES_DIR = pathlib.Path(__file__).resolve().parents[1] / "images"


def data_f32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.random((car, dim)).astype(numpy.float32)


def data_f64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.random((car, dim)).astype(numpy.float64)


def data_i32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    data = rng.integers(0, 2, (car, dim)).astype(numpy.int32)
    data[data == 0] = -1
    return data


def data_i64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    data = rng.integers(0, 2, (car, dim)).astype(numpy.int64)
    data[data == 0] = -1
    return data


def data_u32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.integers(0, 2, (car, dim)).astype(numpy.uint32)


def data_u64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.integers(0, 2, (car, dim)).astype(numpy.uint64)


def gen_strings(car: int, dim: int, alphabet: str) -> list[str]:
    """Return a list of random strings."""
    rng = numpy.random.default_rng()
    return ["".join(rng.choice(list(alphabet), dim)) for _ in range(car)]


def make_plot(  # noqa: PLR0913
    *,
    x: list[int],
    y_abd: list[float],
    fn_name: str,
    dt_name: str,
    y_units: str = "µs",
    y_comp: list[float] | None = None,
    x_label: str = "Dimension",
    y_comp_label: str = "SciPy",
) -> None:
    """Create and save a plot."""
    seaborn.set_theme(style="whitegrid", rc={"figure.figsize": (8, 5)})

    # Add lines with dots for markers
    plot = seaborn.lineplot(x=x, y=y_abd, marker="o", label="ABD")
    if y_comp is not None:
        plot = seaborn.lineplot(x=x, y=y_comp, marker="o", label=y_comp_label)

    # Set the axis labels.
    # The x-axis is the dimension and the y-axis is the time.
    plot.set(xlabel=x_label, ylabel=f"Time per distance call ({y_units})")

    # Set the title
    plot.set_title(f"{fn_name} ({dt_name})")

    # Set a tight layout to remove the whitespace around the plot
    plot.figure.tight_layout()

    # Save with 300 DPI
    plot.figure.savefig(IMAGES_DIR / f"{fn_name}_{dt_name}.png", dpi=200)
    plot.figure.clf()



"""Distance functions for strings."""

from .abd_distances import strings as abd_strings

hamming = abd_strings.hamming
levenshtein = abd_strings.levenshtein
needleman_wunsch = abd_strings.needleman_wunsch


__all__ = [
    "hamming",
    "levenshtein",
    "needleman_wunsch",
]
